@@ -1,6 +1,6 @@
 package PDL::NDBin;
 # ABSTRACT: Multidimensional binning & histogramming
-$PDL::NDBin::VERSION = '0.016';
+$PDL::NDBin::VERSION = '0.017';
 use strict;
 use warnings;
 use Exporter;
@@ -10,6 +10,7 @@ use Math::Round qw( nlowmult nhimult );
 use PDL::Lite;		# do not import any functions into this namespace
 use PDL::NDBin::Iterator;
 use PDL::NDBin::Actions_PP;
+use PDL::NDBin::Utils_PP;
 use Log::Any qw( $log );
 use Data::Dumper;
 use UUID::Tiny qw( :std );
@@ -28,6 +29,32 @@ our %EXPORT_TAGS = ( all => [ qw( ndbinning ndbin ) ] );
 my %valid_key = map { $_ => 1 } qw( axes vars );
 
 
+my @axis_params = qw( max min n round step grid );
+my ( %axis_params, %axis_flags );
+@axis_params{@axis_params} = (1) x @axis_params;
+@axis_flags{@axis_params} = map { 1<<$_ } 0..@axis_params-1;
+
+my %axis_allowed =
+    map { reduce( sub { $a | $b }, 0, @axis_flags{@$_} ) => 1 }
+        [ ],
+	[ qw( n ) ],
+	[ qw( min ) ],
+	[ qw( max ) ],
+        [ qw( step ) ],
+	[ qw( min step ) ],
+	[ qw( max step ) ],
+	[ qw( min n ) ],
+	[ qw( max n ) ],
+	[ qw( round n ) ],
+	[ qw( round step ) ],
+	[ qw( n step ) ],
+	[ qw( n step round ) ],
+	[ qw( min n step ) ],
+	[ qw( max n step ) ],
+	[ qw( min max n ) ],
+	[ qw( min max step ) ],
+        [ qw( grid ) ];
+
 sub add_axis
 {
 	my $self = shift;
@@ -39,8 +66,14 @@ sub add_axis
 			pdl   => 0,
 			round => 0,
 			step  => 0,
+			grid  => 0,
 		} );
 	$log->tracef( 'adding axis with specs %s', \%params );
+
+	my $pmask = reduce { $a | ($b||0) } 0, @axis_flags{ keys %params };
+	croak( "inconsistent or incomplete parameters: ", keys %params )
+	    unless $axis_allowed{ $pmask };
+
 	push @{ $self->{axes} }, \%params;
 }
 
@@ -200,6 +233,21 @@ sub autoscale_axis
 		$axis->{n} = 0;
 		return;
 	}
+
+	# return early if a grid has been supplied
+	if( defined $axis->{grid} ) {
+
+	        $axis->{grid} = PDL::Core::topdl( $axis->{grid} );
+	        croak( "grid supplied for %s must be one-dimensional with at least two elements", $axis )
+		  if $axis->{grid}->nelem < 2 || $axis->{grid}->ndims > 1;
+		_validate_grid( $axis->{grid} );
+		# number of bins is one less than number of bin edges
+		$axis->{n} = $axis->{grid}->nelem - 1;
+		$log->tracef( 'grid supplied for %s; no need to autoscale', $axis );
+		return;
+	}
+
+
 	$axis->{min} = $axis->{pdl}->min unless defined $axis->{min};
 
 
@@ -277,17 +325,33 @@ sub labels
 	$self->autoscale( @_ );
 	my @list = map {
 		my $axis = $_;
-		my ( $pdl, $min, $step ) = @{ $axis }{ qw( pdl min step ) };
-		[ map {
-			{ # anonymous hash
-				range => $pdl->type() >= PDL::float()
-					? [ $min + $step*$_, $min + $step*($_+1) ]
-					: $step > 1
-						? [ nhimult( 1, $min + $step*$_ ), nhimult( 1, $min + $step*($_+1) - 1 ) ]
-						: $min + $step*$_
-			}
-		} 0 .. $axis->{n}-1 ];
+
+		if ( defined $axis->{grid} ) {
+
+		     [ map {
+			      { range => [ $axis->{grid}->at($_), $axis->{grid}->at($_+1) ] }
+		           } 0..$axis->{grid}->nelem -2
+		     ]
+
+		}
+
+		else {
+
+			my ( $pdl, $min, $step ) = @{ $axis }{ qw( pdl min step ) };
+			[ map {
+				{ # anonymous hash
+					range => $pdl->type() >= PDL::float()
+						? [ $min + $step*$_, $min + $step*($_+1) ]
+						: $step > 1
+							? [ nhimult( 1, $min + $step*$_ ), nhimult( 1, $min + $step*($_+1) - 1 ) ]
+							: $min + $step*$_
+				}
+			} 0 .. $axis->{n}-1 ];
+
+	    }
+
 	} $self->axes;
+
 	return wantarray ? @list : \@list;
 }
 
@@ -310,12 +374,23 @@ sub process
 	# find the last axis and flatten all axes into one dimension, working
 	# our way backwards from the last to the first axis
 	for my $axis ( reverse $self->axes ) {
-		$log->debug( 'input (' . $axis->{pdl}->info . ') = ' . $axis->{pdl} ) if $log->is_debug;
-		$log->debug( "bin with parameters step=$axis->{step}, min=$axis->{min}, n=$axis->{n}" )
-			if $log->is_debug;
+	        if ( $log->is_debug ) {
+		    $log->debug( 'input (' . $axis->{pdl}->info . ') = ' . $axis->{pdl} );
+		    if ( ! defined $axis->{grid} ) {
+			$log->debug( "bin with parameters step=$axis->{step}, min=$axis->{min}, n=$axis->{n}" );
+		    }
+		    else {
+			$log->debug( "bin with parameters grid=$axis->{grid}" );
+		    }
+		}
 		croak( 'I cannot bin unless n > 0' ) unless $axis->{n} > 0;
 		unshift @n, $axis->{n};			# remember that we are working backwards!
-		$idx = $axis->{pdl}->_flatten_into( $idx, $axis->{step}, $axis->{min}, $axis->{n} );
+		if ( defined $axis->{grid} ) {
+		    $idx = $axis->{pdl}->_flatten_into_grid( $idx, $axis->{grid} );
+		}
+		else {
+		    $idx = $axis->{pdl}->_flatten_into( $idx, $axis->{step}, $axis->{min}, $axis->{n} );
+		}
 	}
 	$log->debug( 'idx (' . $idx->info . ') = ' . $idx ) if $log->is_debug;
 	$self->{n} = \@n;
@@ -528,7 +603,7 @@ PDL::NDBin - Multidimensional binning & histogramming
 
 =head1 VERSION
 
-version 0.016
+version 0.017
 
 =head1 SYNOPSIS
 
@@ -694,6 +769,21 @@ mandatory.
 
 	$self->add_axis( name => 'longitude', min => -70, max => 70, n => 14 );
 
+Axes may be binned either on a uniformly spaced grid or on a user-provided grid.
+If no specifications are given, data will be binned on a uniformly spaced grid
+automatically derived from the data.
+
+Uniformly spaced grids are specified via a subset of the C<min>,
+C<max>, C<step>, C<n>, and C<round> parameters, while the user
+provided grid is specified via the C<grid> parameter. Only specify the
+subset of parameters which exactly determines the grid.  For example:
+
+  min max n
+  min step n
+  max step n
+  step n round
+  grid
+
 The following axis specifications are available:
 
 =over 4
@@ -730,10 +820,24 @@ to the number of data values, or to 100, whichever is smaller.
 
 Round I<min> and I<max> to the nearest multiple of this value.
 
-=back
+=item grid
 
-Note that you cannot specify all specifications at the same time, because some
-may conflict.
+Either a piddle or a reference to an array containing the values of the
+bin I<boundaries>.  The first and lest elements specify the minimum
+and maximum bin values, while the intermediate elements specify the common
+boundaries.  A bin is I<inclusive> of its lower bound
+and I<exclusive> of its upper bound. For example, the following set of
+bins
+
+  0 <= v < 1
+  1 <= v < 3
+  3 <= v < 9
+
+is represented via
+
+  grid => [ 0, 1, 3, 9 ]
+
+=back
 
 =head2 add_var()
 
@@ -1447,6 +1551,9 @@ even if the bin is empty and not processed further.
 
 =head2 Automatic axis parameter calculation
 
+(Note that if the user defines a binning scheme via the C<grid>
+parameter, no axis parameter calculations are performed.)
+
 =head3 Range
 
 The range, when not given explicitly, is calculated from the data by calling
@@ -1639,14 +1746,14 @@ PDL::NDBin and related distributions on CPAN:
 	| Performance                                       | Low     | Medium | High   | Very high | High     |
 	| Support for weighted histograms                   | X       | X      | X      | X         | -        |
 	| Uses PDL threading                                | -       | -      | -      | X         | -        |
-	| Variable-width bins                               | X       | X      | X      | -         | -        |
+	| Variable-width bins                               | X       | X      | X      | -         | X        |
 	+---------------------------------------------------+---------+--------+--------+-----------+----------+
 
 	  MGH   = Math::GSL 0.26 (Math::GSL::Histogram and Math::GSL::Histogram2D)
 	  MH    = Math::Histogram 1.03
 	  MSHXS = Math::SimpleHisto::XS 1.28
 	  PDL   = PDL 2.4.11
-	  PND   = PDL::NDBin 0.008
+	  PND   = PDL::NDBin 0.017
 
 An explanation and discussion of each of the features is provided below.
 
@@ -1845,8 +1952,11 @@ does not (yet) support threading.
 In a histogram with variable-width bins, the width of the bins needn't be
 equal. This feature can be useful, for example, to construct bins on a
 logarithmic scale. Math::GSL, Math::Histogram, and Math::SimpleHisto::XS
-support variable-width bins; PDL and PDL::NDBin do not and are limited to
-fixed-width bins.
+support variable-width bins; PDL does not, and is limited to fixed-width bins.
+
+Since version 0.017, PDL::NDBin supports variable-width bins if a piddle or
+Perl array containing the bin boundaries is passed in via the I<grid> parameter
+to axis specifications.
 
 =back
 
@@ -2095,8 +2205,7 @@ bin as a list or piddle; this would be very useful for plotting or output.
 
 =item *
 
-PDL::NDBin does not currently support variable-width bins and weighted
-histograms.
+PDL::NDBin does not currently support weighted histograms.
 
 =item *
 
@@ -2109,13 +2218,25 @@ methods such as labels(), n(), step(), etc.
 
 =back
 
+=head1 ACKNOWLEDGEMENTS
+
+=over 4
+
+=item *
+
+Diab Jerius <djerius@cfa.harvard.edu> implemented support for passing in a
+user-defined piddle or array containing bin boundaries, effectively allowing
+variable-width bins. Thanks Diab!
+
+=back
+
 =head1 AUTHOR
 
 Edward Baudrez <ebaudrez@cpan.org>
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2014 by Edward Baudrez.
+This software is copyright (c) 2015 by Edward Baudrez.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
